// https://doc.rust-lang.org/src/std/io/mod.rs.html

use crate::CifError;
use std::io::{Error, Read};

pub trait UnsafeBufRead: Read {
    fn consume(&mut self, amt: usize);
    fn fill_buf(&mut self) -> Result<&[u8], Error>;
    fn read_until(&mut self, byte: u8, buf: &mut Vec<u8>) -> Result<usize, CifError> {
        read_until(self, byte, buf)
    }

    fn read_line(&mut self, buf: &mut Vec<u8>) -> usize {
        read_until(self, b'\n', buf).unwrap()
    }

    fn lines(self) -> UnsafeLines<Self>
    where
        Self: Sized,
    {
        UnsafeLines { buf: self }
    }
}

#[derive(Debug)]
pub struct UnsafeLines<B> {
    buf: B,
}

impl<B: UnsafeBufRead> Iterator for UnsafeLines<B> {
    type Item = Result<String, CifError>;

    fn next(&mut self) -> Option<Result<String, CifError>> {
        let mut buf = Vec::new();
        match self.buf.read_line(&mut buf) {
            0 => None,
            _n => {
                match buf[buf.len() - 1] {
                    b'\n' => {
                        let _ = buf.pop();
                    }
                    _ => {}
                }
                unsafe { Some(Ok(String::from(std::str::from_utf8_unchecked(&buf)))) }
            }
        }
    }
}

fn read_until<R: UnsafeBufRead + ?Sized>(r: &mut R, delim: u8, buf: &mut Vec<u8>) -> Result<usize, CifError> {
    let mut read = 0;
    loop {
        let (done, used) = {
            let available = match r.fill_buf() {
                Ok(n) => n,
                Err(ref e) if e.kind() == std::io::ErrorKind::Interrupted => continue,
                Err(e) => return Err(CifError::IoError(e)),
            };
            match core::slice::memchr::memchr(delim, available) {
                Some(i) => {
                    buf.extend_from_slice(&available[..=i]);
                    (true, i + 1)
                }
                None => {
                    buf.extend_from_slice(available);
                    (false, available.len())
                }
            }
        };
        r.consume(used);
        read += used;
        if done || used == 0 {
            return Ok(read)
        }
    }
}
